# pylint: disable=singleton-comparison
"""Tests pipeline functionality."""
from test.util import get_setup

import pytest
import redis as redis_lib


@pytest.mark.parametrize("rt_lua", [False, True])
def test_pipe(rt_lua: bool) -> None:
    """
    Tests pipeline functionality.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_pipe", rt_lua)

    rt.rpush("foo", "a", "b", "c", "d")
    rt.rpush("bar", "e", "f", "g")
    with rt.pipeline() as pipe:
        pipe.lpop("foo", 3)
        assert rt.llen("foo") == 4
        rt.lpush("foo", "-")
        assert rt.llen("foo") == 5
        pipe.rpop("bar")
        assert rt.llen("bar") == 3
        pipe.rpush("baz", "h")
        assert rt.llen("baz") == 0
        lpop_foo, lpop_bar, rpush_baz = pipe.execute()
    assert lpop_foo == ["-", "a", "b"]
    assert lpop_bar == "g"
    assert rpush_baz == 1
    assert rt.lpop("foo", 2) == ["c", "d"]
    assert rt.rpop("bar") == "f"
    assert rt.rpush("baz", "i") == 2

    with rt.pipeline() as pipe:
        assert rt.zadd("zset", {
            "a": 5,
            "b": 6,
            "c": 4,
        }) == 3
        pipe.zadd("zset", {
            "a": 0,
            "b": -1,
            "c": 1,
        })
        assert rt.zpop_min("zset") == [("c", 4)]
        pipe.zpop_min("zset")
        assert rt.zpop_max("zset") == [("b", 6)]
        pipe.zpop_max("zset")
        assert rt.zcard("zset") == 1
        pipe.zcard("zset")
        zadd_zset, zmin_zset, zmax_zset, zcard_zset = pipe.execute()
    assert zadd_zset == 2
    assert zmin_zset == [("b", -1)]
    assert zmax_zset == [("c", 1)]
    assert zcard_zset == 1

    rt.set("value", "5")
    rt.set("third", "3")
    rt.lpush("lval", "1")
    rt.hset("hval", {"b": "b"})
    rt.zadd("zval", {"c": 1.0})
    rt.lpush("cval", "abc")
    assert rt.exists("value", "third", "lval", "hval", "zval") == 5
    with rt.pipeline() as pipe:
        pipe.delete("value")
        pipe.exists("value")
        pipe.set("value", "10")
        pipe.exists("value")
        v_0, v_1, v_2, v_3 = pipe.execute()
        pipe.set("other", "a")
        pipe.delete("third")
        v_4, v_5 = pipe.execute()
        pipe.delete("lval")
        pipe.exists("lval")
        pipe.lpush("lval", "2")
        v_6, v_7, v_8 = pipe.execute()
        pipe.delete("hval")
        pipe.exists("hval")
        pipe.hset("hval", {"a": "a"})
        v_9, v_10, v_11 = pipe.execute()
        pipe.delete("zval")
        pipe.exists("zval")
        pipe.zadd("zval", {"a": 0.5, "b": 1.5})
        v_12, v_13, v_14 = pipe.execute()
        pipe.delete("cval")
        pipe.exists("cval")
        pipe.hset("cval", {"a": "0", "b": "1", "c": "2"})
        v_15, v_16, v_17 = pipe.execute()
        pipe.set("late_val", "a")
        pipe.delete("late_val")
        pipe.set("late_val", "b")
        pipe.delete("late_val")
        pipe.set("late_val", "c")
        v_18, v_19, v_20, v_21, v_22 = pipe.execute()
    assert v_0 == True  # noqa
    assert v_1 == 0
    assert v_2 == True  # noqa
    assert v_3 == 1
    assert v_4 == True  # noqa
    assert v_5 == True  # noqa
    assert v_6 == True  # noqa
    assert v_7 == 0
    assert v_8 == 1
    assert v_9 == True  # noqa
    assert v_10 == 0
    assert v_11 == 1
    assert v_12 == True  # noqa
    assert v_13 == 0
    assert v_14 == 2
    assert v_15 == True  # noqa
    assert v_16 == 0
    assert v_17 == 3
    assert v_18 == True  # noqa
    assert v_19 == True  # noqa
    assert v_20 == True  # noqa
    assert v_21 == True  # noqa
    assert v_22 == True  # noqa
    assert rt.exists("value") == 1
    assert rt.exists("other") == 1
    assert rt.exists("third") == 0
    assert rt.get("value") == "10"
    assert rt.get("other") == "a"
    assert rt.get("third") is None
    assert rt.lpop("lval") == "2"
    assert rt.hgetall("hval") == {"a": "a"}
    assert rt.zpop_max("zval", 3) == [("b", 1.5), ("a", 0.5)]
    assert rt.hgetall("cval") == {"a": "0", "b": "1", "c": "2"}
    assert rt.hkeys("cval") == ["a", "b", "c"]
    assert rt.hvals("cval") == ["0", "1", "2"]
    if rt_lua:
        with pytest.raises(
                redis_lib.exceptions.ResponseError,
                match=(
                    "WRONGTYPE Operation against a key holding "
                    "the wrong kind of value")):
            assert rt.lpop("cval")
    else:
        with pytest.raises(ValueError, match="key cval already used as hash"):
            assert rt.lpop("cval")
    assert rt.get("late_val") == "c"



"""General functions for scripts."""
from typing import Literal

from redipy.graph.expr import ExprObj
from redipy.symbolic.expr import Constant, Expr, lit_helper, MixedType


class CallFn(Expr):
    """Calls a function. The name of the function must be known at compile
    time."""
    def __init__(
            self,
            fname: str,
            *args: MixedType,
            no_adjust: bool = False) -> None:
        """
        Creates a function call.

        Args:
            fname (str): The name of the function.

            *args (MixedType): The arguments of the function.

            no_adjust (bool, optional): Whether to disallow patching the
            function call. In most cases this should be False.
            Defaults to False.
        """
        self._fname = fname
        self._args: list[Expr] = [lit_helper(arg) for arg in args]
        self._no_adjust = no_adjust

    def compile(self) -> ExprObj:
        return {
            "kind": "call",
            "name": self._fname,
            "args": [arg.compile() for arg in self._args],
            "no_adjust": self._no_adjust,
        }


class FindFn(CallFn):
    """Finds a substring inside a string."""
    def __init__(
            self,
            haystack: MixedType,
            needle: MixedType,
            start_ix: MixedType = None) -> None:
        """
        Finds a substring inside a string.

        Args:
            haystack (MixedType): The string to search in.

            needle (MixedType): The string to search for.

            start_ix (MixedType, optional): The starting index of the search.
            Defaults to None.
        """
        super().__init__(
            "string.find",
            haystack,
            needle,
            *([] if start_ix is None else [start_ix]))


class FromJSON(CallFn):
    """Decodes a string into a JSON object."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("cjson.decode", arg)


class ToJSON(CallFn):
    """Encodes a JSON object into a string."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("cjson.encode", arg)


class ToNum(CallFn):
    """Converts the expression into a number."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("tonumber", arg)


class ToIntStr(CallFn):
    """Converts a number into an integer string representation."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("asintstr", arg)


class ToStr(CallFn):
    """Converts the expression into a string. If you need an integer string for
    a number use `ToIntStr` instead."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("tostring", arg)


class TypeStr(CallFn):
    """Converts the expression into the name of its type."""
    def __init__(self, arg: MixedType) -> None:
        super().__init__("type", arg)


class RedisFn(CallFn):
    """Calls a redis function. The name of the redis function must be known
    at compile time."""
    def __init__(
            self,
            redis_fn: str,
            key: MixedType,
            *args: MixedType,
            no_adjust: bool = False) -> None:
        """
        Creates a function call.

        Args:
            redis_fn (str): The name of the redis function.

            key (MixedType): The key.

            *args (MixedType): The arguments of the redis function.

            no_adjust (bool, optional): Whether to disallow patching the
            function call. In most cases this should be False.
            Defaults to False.
        """
        super().__init__(
            "redis.call", redis_fn, key, *args, no_adjust=no_adjust)


class RedisObj:
    """An object to operate redis functions on a key. Subclassing this class
    is the most common way of exposing redis functionality."""
    def __init__(self, key: MixedType) -> None:
        """
        Creates a redis object for the given key.

        Args:
            key (MixedType): The key.
        """
        self._key = lit_helper(key)

    def key(self) -> Expr:
        """
        Returns the key expression.

        Returns:
            Expr: The key expression.
        """
        return self._key

    def redis_fn(
            self,
            name: str,
            *args: MixedType,
            no_adjust: bool = False) -> Expr:
        """
        Calls a redis function with the given name, key, and arguments.

        Args:
            name (str): The name of the redis function.

            *args (MixedType): The remaining arguments of the redis function.

            no_adjust (bool, optional): Whether to disallow patching the
            function call. In most cases this should be False.
            Defaults to False.

        Returns:
            Expr: The function call.
        """
        return RedisFn(name, self.key(), *args, no_adjust=no_adjust)

    def exists(self) -> Expr:
        """
        Computes whether the key exists calling the `EXISTS` redis function.

        Returns:
            Expr: The expression for the redis function call.
        """
        return self.redis_fn("exists")

    def delete(self) -> Expr:
        """
        Deletes the key calling the `DEL` redis function.

        Returns:
            Expr: The expression for the redis function call.
        """
        return self.redis_fn("del")


LogLevel = Literal["debug", "verbose", "notice", "warning"]
"""Logging level constants."""


class LogFn(CallFn):
    """Logs a string to the default output depending on the logging level."""
    def __init__(
            self,
            level: LogLevel,
            message: MixedType) -> None:
        """
        Logs a string to the default output depending on the logging level.

        Args:
            level (LogLevel): The logging level.

            message (MixedType): The string.
        """
        super().__init__("redis.log", self.convert_level(level), message)

    @staticmethod
    def convert_level(
            level: LogLevel) -> Expr:
        """
        Converts a logging level to constants.

        Args:
            level (LogLevel): The logging level.

        Raises:
            ValueError: If the logging level is invalid.

        Returns:
            Expr: The corresponding constant.
        """
        level_map: dict[LogLevel, str] = {
            "debug": "redis.LOG_DEBUG",
            "verbose": "redis.LOG_VERBOSE",
            "notice": "redis.LOG_NOTICE",
            "warning": "redis.LOG_WARNING",
        }
        raw = level_map.get(level)
        if raw is None:
            raise ValueError(f"could not determine debug level: {level}")
        return Constant(raw)



"""This module offers functionality to create Redis functions. The `FnContext`
class is the base of all scripts and after defining the script it can be
registered via the `register_script` method of `redipy.Redis`.

This module re-exports various scripting functionality for convenience. The
module exists as unified module for all scripting related symbols.
"""
from redipy.backend.backend import ExecFunction
from redipy.graph.expr import JSONType
from redipy.symbolic.expr import Constant, Strs
from redipy.symbolic.fun import (
    CallFn,
    FindFn,
    FromJSON,
    LogFn,
    LogLevel,
    RedisFn,
    ToIntStr,
    ToJSON,
    ToNum,
    ToStr,
    TypeStr,
)
from redipy.symbolic.rhash import RedisHash
from redipy.symbolic.rlist import RedisList
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.rzset import RedisSortedSet
from redipy.symbolic.seq import FnContext


__all__ = [
    "CallFn",
    "Constant",
    "ExecFunction",
    "FindFn",
    "FnContext",
    "FromJSON",
    "JSONType",
    "LogFn",
    "LogLevel",
    "RedisFn",
    "RedisHash",
    "RedisList",
    "RedisSortedSet",
    "RedisVar",
    "Strs",
    "ToIntStr",
    "ToJSON",
    "ToNum",
    "ToStr",
    "TypeStr",
]



"""Implements redis functionality for the memory backend."""
from typing import cast

from redipy.api import RSetMode, RSM_ALWAYS, RSM_EXISTS, RSM_MISSING
from redipy.graph.expr import JSONType
from redipy.memory.state import Machine
from redipy.plugin import ArgcSpec, LocalRedisFunction


class RExistsFn(LocalRedisFunction):
    """Implements the exists function."""
    @staticmethod
    def name() -> str:
        return "exists"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.exists(key)


class RDelFn(LocalRedisFunction):
    """Implements the del function."""
    @staticmethod
    def name() -> str:
        return "del"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.delete(key)


class RSetFn(LocalRedisFunction):
    """Implements the set function."""
    @staticmethod
    def name() -> str:
        return "set"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        value = f"{args[0]}"
        args = args[1:]
        mode: RSetMode = RSM_ALWAYS
        return_previous = False
        expire_in = None
        keep_ttl = False
        pos = 0
        while pos < len(args):
            arg = f"{args[pos]}".upper()
            if arg == "XX":
                mode = RSM_EXISTS
            elif arg == "NX":
                mode = RSM_MISSING
            elif arg == "GET":
                return_previous = True
            elif arg == "PX":
                pos += 1
                expire_in = float(cast(float, args[pos])) / 1000.0
            elif arg == "KEEPTTL":
                keep_ttl = True
            pos += 1
        return sm.set(
            key,
            value,
            mode=mode,
            return_previous=return_previous,
            expire_in=expire_in,
            keep_ttl=keep_ttl)


class RGetFn(LocalRedisFunction):
    """Implements the get function."""
    @staticmethod
    def name() -> str:
        return "get"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.get(key)


class RIncrByFn(LocalRedisFunction):
    """Implements the incrby and incrbyfloat functions."""
    @staticmethod
    def name() -> str:
        return "incrby"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.incrby(key, float(cast(float, args[0])))


class RLPushFn(LocalRedisFunction):
    """Implements lpush."""
    @staticmethod
    def name() -> str:
        return "lpush"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.lpush(key, *(f"{arg}" for arg in args))


class RRPushFn(LocalRedisFunction):
    """Implements rpush."""
    @staticmethod
    def name() -> str:
        return "rpush"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.rpush(key, *(f"{arg}" for arg in args))


class RLPopFn(LocalRedisFunction):
    """Implements lpop."""
    @staticmethod
    def name() -> str:
        return "lpop"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
            "at_most": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.lpop(
            key, None if len(args) < 1 else int(cast(int, args[0])))


class RRPopFn(LocalRedisFunction):
    """Implements rpop."""
    @staticmethod
    def name() -> str:
        return "rpop"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
            "at_most": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.rpop(
            key, None if len(args) < 1 else int(cast(int, args[0])))


class RLRangeFn(LocalRedisFunction):
    """Implements the lrange function."""
    @staticmethod
    def name() -> str:
        return "lrange"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.lrange(key, int(cast(int, args[0])), int(cast(int, args[1])))


class RLLenFn(LocalRedisFunction):
    """Implements the llen function."""
    @staticmethod
    def name() -> str:
        return "llen"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.llen(key)


class RZAddFn(LocalRedisFunction):
    """Implements the zadd function."""
    @staticmethod
    def name() -> str:
        return "zadd"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.zadd(key, {f"{args[1]}": float(cast(float, args[0]))})


class RZPopMaxFn(LocalRedisFunction):
    """Implements zpopmax."""
    @staticmethod
    def name() -> str:
        return "zpopmax"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
            "at_most": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return cast(list | None, sm.zpop_max(
            key, 1 if len(args) < 1 else int(cast(int, args[0]))))


class RZPopMinFn(LocalRedisFunction):
    """Implements zpopmin."""
    @staticmethod
    def name() -> str:
        return "zpopmin"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
            "at_most": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return cast(list | None, sm.zpop_min(
            key, 1 if len(args) < 1 else int(cast(int, args[0]))))


class RZRangeFn(LocalRedisFunction):
    """Implements the zrange function."""
    @staticmethod
    def name() -> str:
        return "zrange"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.zrange(key, int(cast(int, args[0])), int(cast(int, args[1])))


class RZCardFn(LocalRedisFunction):
    """Implements the zcard function."""
    @staticmethod
    def name() -> str:
        return "zcard"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.zcard(key)


class RHSetFn(LocalRedisFunction):
    """Implements the hset function."""
    @staticmethod
    def name() -> str:
        return "hset"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        mapping = {}
        ix = 0
        while ix < len(args):
            field = f"{args[ix]}"
            ix += 1
            if ix >= len(args):
                raise ValueError(f"unbalanced field value pairs: {args}")
            value = f"{args[ix]}"
            ix += 1
            mapping[field] = value
        return sm.hset(key, mapping)


class RHDelFn(LocalRedisFunction):
    """Implements the hdel function."""
    @staticmethod
    def name() -> str:
        return "hdel"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hdel(key, *(f"{arg}" for arg in args))


class RHGetFn(LocalRedisFunction):
    """Implements the hget function."""
    @staticmethod
    def name() -> str:
        return "hget"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hget(key, f"{args[0]}")


class RHMGetFn(LocalRedisFunction):
    """Implements the hmget function."""
    @staticmethod
    def name() -> str:
        return "hmget"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
            "at_least": True,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hmget(key, *(f"{arg}" for arg in args))


class RHIncrByFn(LocalRedisFunction):
    """Implements the hincrby and hincrbyfloat functions."""
    @staticmethod
    def name() -> str:
        return "hincrby"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hincrby(key, f"{args[0]}", cast(float, args[1]))


class RHKeysFn(LocalRedisFunction):
    """Implements the hkeys function."""
    @staticmethod
    def name() -> str:
        return "hkeys"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hkeys(key)


class RHValsFn(LocalRedisFunction):
    """Implements the hvals function."""
    @staticmethod
    def name() -> str:
        return "hvals"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hvals(key)


class RHGetAllFn(LocalRedisFunction):
    """Implements the hgetall function."""
    @staticmethod
    def name() -> str:
        return "hgetall"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 0,
        }

    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        return sm.hgetall(key)
